r"""A module with utility functions and classes around and for creating control
problems, agents and their components, and typing.

Submodules
==========

.. autosummary::
   :toctree: generated
   :template: module.rst
   :caption: Submodules

   control
   geometry
   iters
   math
   named
   seeding
"""

__all__ = ["control", "geometry", "iters", "math", "named", "seeding"]

from . import control, geometry, iters, math, named, seeding
